package com.danmassey.rugbe.board;

import java.util.ArrayList;
import java.util.Iterator;

import com.danmassey.rugbe.robot.NBot;

/**
 * Maintains the board layout and information.
 * 
 * @author dan
 * Created on Aug 30, 2002
 */
public class Board {
	public static final char HOME = '@';
	public static final char PLAIN = '.';
	public static final char WALL = '#';
	public static final char WATER = '~';
	public static final char EDGE = '|';
	public static final char ROBOT = '\u263B';
	public static final char BUFFER_1 = '5';
	public static final char BUFFER_2 = '3';
	
	
	private char[] grid;
	private int height;
	private int width;
	private ArrayList homes = new ArrayList(32);
	private ArrayList uncoloredHomes;
	private char[] uncoloredGrid;
	
	
	public Board(int width, int height, char[] grid) {
		this.width = width;
		this.height = height;
		this.grid = grid;
		this.uncoloredGrid = grid;
		initHomes();
	}

	/**
	 * Method initHomes.
	 */
	private void initHomes() {
		for (int i = 0; i < grid.length; i++) {
			if (grid[i] == Board.HOME) {
				homes.add(new Node(gridToX(i), gridToY(i)));
			}
		}
		this.uncoloredHomes = new ArrayList(homes);
	}

	
	public char[] getGrid() {
		return this.grid;
	}
	
	public int getHeight() {
		return this.height;
	}
	
	public int getwidth() {
		return this.width;
	}
	
	public char getTileAt(int x, int y) {
		if (offMap(x, y)) {
			return Board.EDGE;
		}
		return grid[xyToGrid(x, y)];
	}

	/**
	 * Method offMap.
	 * @param x
	 * @param y
	 * @return boolean
	 */
	private boolean offMap(int x, int y) {
		if ((x > width) || (y > height) || (x < 1) || (y < 1)) {
			return true;
		} else {
			return false;
		}
	}

	
	public char getTileNorthOf(int x, int y) {
		return getTileAt(x, y + 1);
	}
	
	public char getTileSouthOf(int x, int y) {
		return getTileAt(x, y - 1);
	}
	
	public char getTileEastOf(int x, int y) {
		return getTileAt(x + 1, y);
	}
	
	public char getTileWestOf(int x, int y) {
		return getTileAt(x - 1, y);
	}
	
	/**
	 * Returns an array of tiles tiles surrounding the current tile. Starts with 
	 * North and runs clockwise.
	 **/
	public char[] getSurroundingTiles(int x, int y) {
		char[] surroundings = new char[4];
		surroundings[0] = getTileNorthOf(x, y);
		surroundings[1] = getTileEastOf(x, y);
		surroundings[2] = getTileSouthOf(x, y);
		surroundings[3] = getTileWestOf(x, y);
		return surroundings;
	}
	
	public Node getClosestHome(int x, int y) {
		int distance = Integer.MAX_VALUE;
		Node node = null;
		Node temp = null;
		int size = homes.size();
		int tempDistance = 0;
		for (int i = 0; i < size; i++ ) {
			temp = (Node) homes.get(i);
			tempDistance = Math.abs(x - temp.getX()) + Math.abs(y - temp.getY());
			if (tempDistance < distance) {
				distance = tempDistance;
				node = temp;
			}			
		}
		return node;
	}
	
	public void removeHome(Node node, boolean empty) {
		this.homes.remove(node);
		if (empty) {
			this.uncoloredHomes.remove(node);
		}
	}
	
	// Factor other bots into the pathing
	public void color(Iterator botIter) {
		while (botIter.hasNext()) {
			NBot bot = (NBot) botIter.next();
			if (!bot.isMe()) {
				int x = bot.getX();
				int y = bot.getY();
				
				// Bot
				grid[xyToGrid(x, y)] = Board.ROBOT;
				
				// One away
				grid[xyToGrid(x, y + 1)] = Board.BUFFER_1; // N
				grid[xyToGrid(x + 1, y)] = Board.BUFFER_1; // E
				grid[xyToGrid(x, y - 1)] = Board.BUFFER_1; // S
				grid[xyToGrid(x - 1, y)] = Board.BUFFER_1; // W
				
				// Two away
				grid[xyToGrid(x, y + 2)] = Board.BUFFER_2;		// N
				grid[xyToGrid(x + 1, y + 1)] = Board.BUFFER_2;	// NE
				grid[xyToGrid(x + 2, y)] = Board.BUFFER_2;		// E
				grid[xyToGrid(x + 1, y - 1)] = Board.BUFFER_2;	// SE
				grid[xyToGrid(x, y - 2)] = Board.BUFFER_2;		// S
				grid[xyToGrid(x - 1, y - 1)] = Board.BUFFER_2;	// SW
				grid[xyToGrid(x - 2, y)] = Board.BUFFER_2;		// W
				grid[xyToGrid(x - 1, y + 1)] = Board.BUFFER_2; 	// NW
			}
		}
	}
	
	public void uncolor() {
		this.grid = new char[this.uncoloredGrid.length];
		System.arraycopy(this.uncoloredGrid, 0, this.grid, 0, this.uncoloredGrid.length);
		this.homes = new ArrayList(this.uncoloredHomes);
	}
	
	public String toString() {
		StringBuffer buffy = new StringBuffer();
		buffy.append(this.width + " " + this.height + "\n");
		for (int i = 0; i < this.grid.length; i++) {
			buffy.append(this.grid[i]);
			if (i % width == (width - 1)) {
				buffy.append("\n");
			}
		}
		buffy.append("\n");
		return buffy.toString();
	}
	
	public int xyToGrid(int x, int y) {
		return (((y - 1) * width) + x) - 1;
	}
	
	public int gridToX(int index) {
		return (index % this.width) + 1;
	}
	
	public int gridToY(int index) {
		return (index / this.width) + 1;
	}
}
