package com.danmassey.rugbe.brain;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.ResourceBundle;

import com.danmassey.rugbe.board.Board;
import com.danmassey.rugbe.board.Node;
import com.danmassey.rugbe.board.Pkg;
import com.danmassey.rugbe.brain.path.AStar;
import com.danmassey.rugbe.net.ComChannel;
import com.danmassey.rugbe.robot.NBot;
import com.danmassey.rugbe.robot.Robot;
import com.danmassey.rugbe.ui.DebugUI;

/**
 * 
 * 
 * @author dan
 * Created on Aug 31, 2002
 **/
public class GoAroundBrain implements IBrain {
	private Robot robot;
	private Board board;
	private LinkedHashMap nbots;
	private ComChannel com;
	private ResourceBundle bundle;
	private LinkedHashMap bases;
	private DebugUI ui;
	private AStar pathFinder;
	private Node nextStep;

	/**
	 * @see com.danmassey.rugbe.brain.IBrain#init(Robot, Board, LinkedHashMap, ComChannel, ResourceBundle)
	 */
	public void init(Robot robot, Board board, LinkedHashMap nbots, ComChannel com,
			ResourceBundle bundle, DebugUI ui) {
		this.robot = robot;
		this.board = board;
		this.nbots = nbots;
		this.com = com;
		this.bundle = bundle;
		this.ui = ui;
		this.pathFinder = new AStar(this.board);
		this.nextStep = null;
	}

	/**
	 * @see com.danmassey.rugbe.brain.IBrain#process(Pkg[])
	 */
	public void process(Pkg[] packages) {
		if (offPath()) {
			nextStep = null; // cleaner to start over
		}
		if (botProximity(4)) {	
			nextStep = null; // cleaner to start over
			board.color(nbots.values().iterator());
		}
		if (deliver()) {
		} else if (pickUp(packages)) {
		} else {
			pathMove();
		} 	
		board.uncolor();
	}

	/**
	 * Method dodge.
	 * @return boolean
	 */
	private boolean botProximity(int p) {
		Iterator iter = nbots.values().iterator();
		NBot bot = null;
		while (iter.hasNext()) {
			bot = (NBot) iter.next();
			if (bot.calcDistance(robot.getX(), robot.getY()) < p) {
				return true;
			}
		}
		return false;
	}
	
	
	private boolean offPath() {
		if (nextStep != null) {
			if (new Node(robot.getX(), robot.getY()).distance(nextStep) > 1) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Method canDeliver.
	 * @return boolean
	 */
	private boolean deliver() {
		boolean deliver = false;
		Pkg pkg = robot.getPackages();
		ArrayList list = new ArrayList(8);
		while (pkg != null) {
			if ((pkg.getDestX() == robot.getX()) && (pkg.getDestY() == robot.getY())) {
				list.add(pkg);
				deliver = true;
			}	
			pkg = pkg.getNextPkg();
		}	
		if (deliver) {
			int[] ids = new int[list.size()];
			for (int i = 0; i < ids.length; i++) {
				ids[i] = ((Pkg) list.get(i)).getId();
			}
			com.drop(ids, 1);
		}	
		return deliver;
	}

	/**
	 * Method canPickUp.
	 * @param packages
	 * @return boolean
	 */
	private boolean pickUp(Pkg[] pkgs) {
		Pkg[] packages = this.sort(pkgs, robot.getX(), robot.getY());
		boolean pickUp = false;
		int[] list = new int[packages.length];
		int remainingCapacity = robot.getCapacity() - robot.getCarrying();
		int j = 0;
		for (int i = 0; i < packages.length; i++) {
			if (packages[i].getWeight() < remainingCapacity) {
				remainingCapacity -= packages[i].getWeight();
				list[j++] = packages[i].getId();
				pickUp = true;		
			}
		}
		if (pickUp) {
			int[] ids = new int[j];
			System.arraycopy(list, 0, ids, 0, j);
			com.pick(ids, 1);
		}
		return pickUp;
	}
	
	private void pathMove() {
		if (nextStep == null) {
			if (robot.getPackages() == null) {
				baseMove();
			} else {
				pkgMove();
			}
		}
		if (nextStep != null) {
			moveCommand(robot.getX(), robot.getY(), nextStep.getX(), nextStep.getY());
			nextStep = nextStep.getChild();
		} else {
			com.drop(new int[] {1}, 1);
		}
			
	}
	
	private void baseMove() {
		while (nextStep == null) {
			Node home = board.getClosestHome(robot.getX(), robot.getY());
			if (home.distance(new Node(robot.getX(), robot.getY())) == 0) {
				board.removeHome(home, true);
				home = board.getClosestHome(robot.getX(), robot.getY());
			}
			if (home == null) {
				break;
			}
			nextStep = pathFinder.findPath(robot.getX(), robot.getY(), home.getX(), home.getY());
			if (nextStep == null) {
				board.removeHome(home, false);
			}
		}
		if ((ui != null) && (nextStep != null)) {
			ui.plotPath(nextStep);
		}
	}
	
	private void pkgMove() {
		while ((nextStep == null) && (robot.getPackages() != null)) {
			Pkg pkg = robot.getPackages();
			nextStep = pathFinder.findPath(robot.getX(), robot.getY(), pkg.getDestX(), pkg.getDestY());
		}
		if ((ui != null) && (nextStep != null)) {
			ui.plotPath(nextStep);
		}
	}

	private void moveCommand(int currentX, int currentY, int newX, int newY) {
		int dX = newX - currentX;
		int dY = newY - currentY;
		
		if (dY == 1) {
			com.move(ComChannel.NORTH, 1);
		} else if (dX == 1) {
			com.move(ComChannel.EAST, 1);
		} else if (dY == -1) {
			com.move(ComChannel.SOUTH, 1);
		} else if (dX == -1) {
			com.move(ComChannel.WEST, 1);
		} else {
			com.drop(new int[] {1}, 1);
		}
	}
	
	private Pkg[] sort(Pkg[] packages, int x, int y) {
		for (int i = 0; i < packages.length; i++) {
			packages[i].setDistance(Math.abs(packages[i].getDestX() - robot.getX())
					+ Math.abs(packages[i].getDestY() - robot.getY()));
		}
		int r;
		Pkg temp;
		for (int q = 1; q < packages.length; q++) {
    		temp = packages[q];
    		r = q;
    		while ((r > 0) && (packages[r-1].getDistance() > temp.getDistance())) {
      			packages[r] = packages[r-1];
      			r = r - 1;
    		}
    		packages[r] = temp;
  		}
 		return packages;
	}
}
